// TomUtil.cpp: implementation of the CTomUtil class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <atlbase.h>
#include "TomUtil.h"
#include "RL2.h"
#import "riched20.dll" named_guids

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif
//DEFINE_GUID(IID_MyGUID,0x05092F20L,0x833F,0x11CF,0xAB,0x59,0x00,0xAA,0x00,0xC0,0x91,0xA1);	

BOOL CTomUtil::DisableInsertMode(const CRichEditCtrl &rich, const BOOL bDisabled)
{
	ITextDocument* const pDoc=GetITextDocument(rich);
	if (pDoc) {
		ITextSelection* pSel=NULL;
		pDoc->GetSelection(&pSel);
		if (pSel) {
			long l=0;
			pSel->GetFlags(&l);
			if (l & tomSelOvertype) {
				l^=tomSelOvertype;
				pSel->SetFlags(l);
			}
			else {
			}
			pSel->Release();
		}
		pDoc->Release();
	}

	return CTomUtil::IsEnabledOVRMode(rich);
}

ITextDocument* CTomUtil::GetITextDocument(const CRichEditCtrl &rich)
{
	return GetITextDocument(rich.GetSafeHwnd());
}

ITextRange* CTomUtil::GetITextRange(const CRichEditCtrl &rich)
{
	CHARRANGE cr;
	cr.cpMin=-1;
	cr.cpMax=-1;
	rich.GetSel(cr);
	return GetITextRange(rich,cr);
}

ITextPara* CTomUtil::GetITextPara(const CRichEditCtrl &rich)
{
	ITextRange* pRange=GetITextRange(rich);
	if (!pRange)
		return NULL;
	
	ITextPara* pPara=NULL;
	pRange->GetPara(&pPara);
	pRange->Release();
	return pPara;
}

ITextRange* CTomUtil::GetSelectedITextRange(const CRichEditCtrl &rich)
{
	CHARRANGE cr;
	rich.GetSel(cr);
	if (cr.cpMin == cr.cpMax)
		return NULL;
	else
		return GetITextRange(rich,cr);
}

bool CTomUtil::IsEnabledOVRMode(const CRichEditCtrl &rich)
{
	bool bRet=true;
	CComPtr<ITextDocument> Doc;
	Doc.Attach(GetITextDocument(rich));
	
	if (Doc) {
		CComPtr<ITextSelection> Sel;
		Doc->GetSelection(&Sel);
		if (Sel) {
			long l=0;			
			Sel->GetFlags(&l);
			
			if (!(l & tomSelOvertype))
				bRet=false;
		}
	}
	return bRet;
}

bool CTomUtil::PasteFileAsWordDocument(const HWND hwndRich, LPCTSTR lpszPathName)
{
	ASSERT(IsWindow(hwndRich));
	ASSERT(!IsBadStringPtr(lpszPathName,INFINITE));

	CComPtr<ITextDocument> Doc;
	Doc.Attach(CTomUtil::GetITextDocument(hwndRich));
	if (Doc) {
		COleVariant var(lpszPathName);
		const HRESULT hr=Doc->Open(var,tomOpenExisting|tomPasteFile,0);
		return (S_OK==hr);
	}
	else
		return false;
}

ITextDocument* CTomUtil::GetITextDocument(const HWND hwnd)
{
	ASSERT(::IsWindow(hwnd));
	IUnknown* pUnk=NULL;
	::SendMessage(hwnd,EM_GETOLEINTERFACE,0,(LPARAM)&pUnk);
	ASSERT(pUnk);
	if (!pUnk)
		return NULL;

	ITextDocument* pDoc=NULL;
	pUnk->QueryInterface(IID_ITextDocument,(LPVOID*)&pDoc);
	pUnk->Release();

	return pDoc;
}

bool CTomUtil::SuspendUndo(ITextDocument *const pDoc, const bool bSuspend)
{
	ASSERT(!IsBadReadPtr(pDoc,sizeof(ITextDocument)));
	if(IsBadReadPtr(pDoc,sizeof(ITextDocument)))
		return false;

	bool bRet=false;
	if (bSuspend) {
		pDoc->Undo(tomSuspend,NULL);
		long l;
		const HRESULT hr=pDoc->Freeze(&l);
		ASSERT(S_OK==hr);
		bRet=(S_OK==hr);
	}
	else {
		long l;
		const HRESULT hr=pDoc->Unfreeze(&l);
		//ASSERT(S_OK==hr);
		pDoc->Undo(tomResume,NULL);
		bRet=true;
	}

	return bRet;
}

bool CTomUtil::SuspendUndo(const HWND hwndRich, const bool bSuspend)
{
	ASSERT(IsWindow(hwndRich));

	ITextDocument* const pDoc=GetITextDocument(hwndRich);
	ASSERT(!IsBadReadPtr(pDoc,sizeof(ITextDocument)));
	if(IsBadReadPtr(pDoc,sizeof(ITextDocument)))
		return false;

	const bool bRet=SuspendUndo(pDoc,bSuspend);
	pDoc->Release();

	return bRet;
}

long GetDirect(const long lStart, const long lEnd)
{
	if (-1==lEnd)
		return tomForward;
	else {
		return (lEnd-lStart);
	}
}

UINT CTomUtil::FindText(const HWND hwndRE
						, LPCTSTR lpszFind
						, const bool bMatchCase
						, const bool bWholeWord
						, const long lStart
						, const long lEnd
						, PFUNC_FINDTEXTPROC pfn
						, const LPARAM lParam)
{

	ASSERT(IsWindow(hwndRE));
	ASSERT(!IsBadStringPtr(lpszFind,INFINITE));
	ASSERT(strlen(lpszFind));
	ASSERT(!IsBadCodePtr((FARPROC)pfn));

	long lFlags=0;
	if (bMatchCase)
		lFlags|=tomMatchCase;
	if (bWholeWord)
		lFlags|=tomMatchWord;

	return FindText(hwndRE,lpszFind,lFlags,lStart,lEnd,pfn,lParam);
}

UINT CTomUtil::FindText(const HWND hwndRE
						, LPCTSTR lpszFind
						, const long lTOMOptions
						, const long lStart
						, const long lEnd
						, PFUNC_FINDTEXTPROC pfn
						, const LPARAM lParam)
{
	ASSERT(IsWindow(hwndRE));
	ASSERT(!IsBadStringPtr(lpszFind,INFINITE));
	ASSERT(strlen(lpszFind));
	ASSERT(!IsBadCodePtr((FARPROC)pfn));

	CComPtr<ITextDocument> Doc;
	Doc.Attach(GetITextDocument(hwndRE));
	ASSERT(Doc);
	if (!Doc)
		return 0;

	CComPtr<ITextRange> Range;
	
	Doc->Range(0==lStart ? tomStart : lStart
		,-1==0 ? tomEnd : lEnd ,&Range);
	
	ASSERT(Range);
	if (!Range)
		return 0;

	//if (lStart)
		//pRange->SetStart(lStart);	
	/*
	      
	         ( 
	*/
	Range->SetStart(lStart);


	CString strFind=CString(lpszFind);
	CComBSTR bstr(strFind.AllocSysString());
	long lLen=0;
	const long lDirect=GetDirect(lStart,lEnd);
	
	UINT nCount=0;
	while (S_OK==Range->FindText(bstr,lDirect,lTOMOptions,&lLen)) {
		nCount++;
		CHARRANGE cr={0};
		VERIFY(S_OK==Range->GetStart(&cr.cpMin));

		if ((lEnd>0) && (cr.cpMin>lEnd))
			break;
		cr.cpMax=cr.cpMin+lLen;
		
		long lFlags=0;
		const bool bNext=pfn(*Range,cr,strFind,lParam,lFlags);
		//  callback    ITextRange::Release
		ASSERT(!IsBadReadPtr(Range,sizeof(ITextRange)));
		if (!bNext)
			break;
		
		if (0 == (lFlags & FNR_NOCORRECTRANGE))
			VERIFY(S_OK==Range->SetStart(cr.cpMax+1));
		if (0 != (lFlags & FNR_STRINGCHANGED)) 
			bstr=strFind.AllocSysString();
	}//while

	//VERIFY(0 == pRange->Release());
	return nCount;
}

ITextRange* CTomUtil::GetITextRange(
	const CRichEditCtrl &rich
	, const CHARRANGE& cr)
{
	return GetITextRange(rich.GetSafeHwnd(),cr);
}

ITextRange* CTomUtil::GetITextRange(const HWND hwnd
	, const CHARRANGE& cr)
{
	ASSERT(IsWindow(hwnd));
	ITextDocument* const pDoc=GetITextDocument(hwnd);
	if (!pDoc)
		return NULL;
	
	ITextRange* pRange=NULL;
	pDoc->Range(cr.cpMin,cr.cpMax,&pRange);
	pDoc->Release();
	return pRange;
}

CString CTomUtil::GetTextFromRange(const HWND hwnd
								   , const long lBegin
								   , const long lEnd)
{
	ASSERT(IsWindow(hwnd));

	//ITextDocument* const pDoc=CTomUtil::GetITextDocument(hwnd);
	CComPtr<ITextDocument> Doc;
	Doc.Attach(CTomUtil::GetITextDocument(hwnd));
	ASSERT(Doc);
	if (!Doc)
		return TEXT("");
	
	CComPtr<ITextRange> Range;
	const HRESULT hr=Doc->Range(lBegin,lEnd,&Range);
	ASSERT(S_OK==hr);
	
	if (S_OK!=hr)
		return TEXT("");

	ASSERT(Range);
	CComBSTR bstr;
	CString strRet;
	if (S_OK==Range->GetText(&bstr)) {
		strRet=CString(bstr);
	}
	return strRet;
}

CString CTomUtil::GetTextFromRange(const HWND hwndRich, const CHARRANGE &cr)
{
	ASSERT(IsWindow(hwndRich));
	return GetTextFromRange(hwndRich,cr.cpMin,cr.cpMax);
}

bool CTomUtil::ReplaceSel(const HWND hwnd, LPCTSTR lpszText)
{
	ASSERT(IsWindow(hwnd));
	CHARRANGE cr;
	SendMessage(hwnd,EM_EXGETSEL,0,(LPARAM)&cr);
	//  
	const int nSelLen=cr.cpMax-cr.cpMin;
	ASSERT(nSelLen>=0);
	const bool bRet=ReplaceText(hwnd,lpszText,cr.cpMin,cr.cpMax);
	if (bRet) {//   
		const CString strText(lpszText);//     NULL==lpszText
		const int nTextLen=strText.GetLength();
		const int delta=nTextLen-nSelLen;
		CHARRANGE crNew;
		crNew.cpMax=cr.cpMax+delta;
		crNew.cpMin=crNew.cpMax;
		SendMessage(hwnd,EM_EXSETSEL,0,(LPARAM)&crNew);
	}

	return bRet;
}

bool CTomUtil::ReplaceText(const HWND hwnd
						   , LPCTSTR lpszText
						   , const long lStart
						   , const long lEnd)
{
	ASSERT(IsWindow(hwnd));
	CComPtr<ITextDocument> Doc;
	Doc.Attach(GetITextDocument(hwnd));
	ASSERT(Doc);
	if (!Doc)
		return false;
	
	CComPtr<ITextRange> R;
	{
	const HRESULT hr=Doc->Range(lStart,lEnd,&R);
	ASSERT(S_OK==hr);
	if (S_OK!=hr)
		return false;
	
	ASSERT(R);
	}

	CComBSTR bstr(lpszText);
	const HRESULT hr=R->SetText(bstr);

	return (S_OK==hr);	
}

bool CTomUtil::SetUnderlineRed(ITextRange *const pRange, const long lUnderlineType)
{
	ASSERT(pRange);
	if (!pRange)
		return false;

	CComPtr<ITextFont> Font;
	if (S_OK!=pRange->GetFont(&Font))
		return false;

	ASSERT(Font);
	const enum {COLOR_RED=5};
	const HRESULT hr=Font->SetUnderline(MAKEWORD(lUnderlineType,COLOR_RED));

	return S_OK==hr;
}

ITextFont* CTomUtil::GetITextFontForRange(const HWND hwnd
									  , const long lStart
									  , const long lEnd)
{
	CHARRANGE cr;
	cr.cpMin=lStart;
	cr.cpMax=lEnd;
	CComPtr<ITextRange> Range;
	Range.Attach(GetITextRange(hwnd,cr));
	if (!Range)
		return NULL;

	ITextFont* pFont=NULL;
	Range->GetFont(&pFont);
	
	return pFont;
}

ITextFont* CTomUtil::GetITextFont(const HWND hwnd)
{
	return GetITextFontForRange(hwnd,0,-1);
}

int CTomUtil::FindTextSimple(const HWND hwndRE, LPCTSTR lpszText, const long lStart)
{
	ASSERT(IsWindow(hwndRE));
	ASSERT(!IsBadStringPtr(lpszText,INFINITE));
	ASSERT(strlen(lpszText));
	
	CComPtr<ITextDocument> Doc;
	Doc.Attach(GetITextDocument(hwndRE));
	ASSERT(Doc);
	if (!Doc)
		return -1;

	CComPtr<ITextRange> Range;
	Doc->Range(lStart, tomEnd,&Range);

	if (lStart>0)
		VERIFY(S_OK == Range->MoveStart(tomCharacter,lStart,NULL));

#ifdef _DEBUG
	{
	long lCheck=-1;
	ASSERT(S_OK==Range->GetStart(&lCheck));
	ASSERT(lCheck == lStart);
	}
#endif
	
	ASSERT(Range);
	if (!Range)
		return -1;

	const CString strFind=CString(lpszText);
	CComBSTR bstr;
	bstr.Attach(strFind.AllocSysString());
	long lLen=0;
	
	const HRESULT hr=Range->FindText(bstr,tomForward,0,&lLen);
	if (S_OK==hr) {
		long l=0;
		Range->GetStart(&l);
		return l;
	}
	else {
		return -1;
	}	
}
/*
static 
long GetTextLengthEx(const HWND hwnd)
{
	ASSERT(hwnd);

	GETTEXTLENGTHEX ss={0};
	ss.codepage=CP_ACP;
	ss.flags=GTL_NUMCHARS|GTL_PRECISE;

	const int nLen=SendMessage(hwnd,EM_GETTEXTLENGTHEX,(WPARAM)&ss,0);
	return nLen;
}*/

bool CTomUtil::SetProtectedForAllText(
	const HWND hwnd, const bool bProtect)
{
	ASSERT(IsWindow(hwnd));
	const bool bRet=SetProtectedForAllTextByAPI(hwnd,bProtect);
	ASSERT(true==bRet);
	return bRet;
/*	
	CComPtr<ITextDocument> Doc=GetITextDocument(hwnd);
	if (!Doc)
		return false;

	CComPtr<ITextRange> Range;

	const long lLen=GetTextLengthEx(hwnd);
	HRESULT hr;
	hr=Doc->Range(0,lLen+2,&Range);
	if (S_OK!=hr)
		hr=Doc->Range(0,lLen+1,&Range);
	if (S_OK!=hr)
		hr=Doc->Range(0,lLen,&Range);
	
	if (S_OK!=hr) {
		ASSERT(FALSE);
		return SetProtectedForAllTextByAPI(hwnd,bProtect);
	}

	CComPtr<ITextFont> Font;
	hr=Range->GetFont(&Font);
	if (S_OK!=hr) {
		ASSERT(FALSE);
		return SetProtectedForAllTextByAPI(hwnd,bProtect);
	}

	hr=Font->SetProtected(bProtect ? tomTrue : tomFalse);
	return (S_OK == hr);*/
}

bool CTomUtil::SetProtectedForAllTextByAPI(const HWND hwnd, const bool bProtect)
{
	ASSERT(IsWindow(hwnd));
	
	CHARRANGE crPrev;
	SendMessage(hwnd,EM_EXGETSEL,0,(LPARAM)&crPrev);
	{
		CHARRANGE cr;
		cr.cpMin=0; cr.cpMax=-1;
		SendMessage(hwnd,EM_EXSETSEL,0,(LPARAM)&cr);
	}
	
	CHARFORMAT2 cf;
	cf.cbSize=sizeof(cf);
	cf.dwMask=CFM_PROTECTED;
	if (bProtect)
		cf.dwEffects=CFE_PROTECTED;

	SendMessage(hwnd,EM_SETCHARFORMAT,SCF_ALL|SCF_SELECTION,(LPARAM)&cf);
	SendMessage(hwnd,EM_EXSETSEL,0,(LPARAM)&crPrev);
	return true;
}

bool CTomUtil::SetHiddenText(const HWND hwnd
							 , const bool bHidden
							 , const long lStart
							 , const long lEnd)
{
	ASSERT(IsWindow(hwnd));

	CComPtr<ITextDocument> Doc;
	Doc.Attach(GetITextDocument(hwnd));
	ASSERT(Doc);
	if (!Doc)
		return false;

	CComPtr<ITextRange> R;
	{//block begin
	const HRESULT hr=Doc->Range(lStart,lEnd,&R);
	if (S_OK!=hr) {
		return false;
	}
	}//block end

	CComPtr<ITextFont> F;
	{//block begin
	const HRESULT hr=R->GetFont(&F);
	if (S_OK!=hr) {
		return false;
	}
	}//block end

	RL2::CReadOnly read_only_locker(hwnd);
	const HRESULT hr=F->SetHidden(bHidden ? tomTrue : tomFalse);
	if (S_OK !=hr) {
		const bool b=E_ACCESSDENIED ==hr;
		const bool b2=b;
	}
	return (S_OK==hr);
}

